import click

from boiler import cli
from boiler.commands.utils import exit_with, handle_request_error
from boiler.definitions import all_activity_codes


@click.group(name='activity', short_help='get information about activities')
@click.pass_obj
def activity(ctx):
    pass


@activity.command(name='list-types', help='list known activities')
@click.option('--count', '-c', is_flag=True, help='print the number of activity types')
def list_activities(count):
    if count:
        click.echo(len(all_activity_codes.values()))
    else:
        for activity_type in sorted(str(t.value) for t in all_activity_codes.values()):
            click.echo(activity_type)


@activity.command(name='list', help='list activities for video')
@click.option('--video-id', required=True)
@click.pass_obj
def list_video_activities(ctx, video_id):
    r = ctx['session'].get(f'video/{video_id}/activities')
    exit_with(handle_request_error(r))


@activity.command(name='count', short_help='get the number of activities by type')
@click.argument('output', type=click.File('wb'), default='-')
@click.pass_obj
def get_activity_count(ctx, output):
    """
    Get a summary of annotated activities grouped by phase.  By default,
    this will print to the terminal.  Add a path argument to write to
    a file.
    """
    r = ctx['session'].get('activity/summary')
    output.write(r.content)


cli.add_command(activity)



import webbrowser

import click
import gitlab
from gitlab.exceptions import GitlabAuthenticationError
import requests

from boiler import BOILER_CONFIG_FILE, cli, update_config_value


def _is_valid_stumpf_token(stumpf_url: str, token: str) -> bool:
    # note - don't use the boiler session since that could have an old token,
    # and automatically fails on auth failures.
    return requests.get(stumpf_url + '/api/v1/user/me', headers={'X-Stumpf-Token': token}).ok


def _is_valid_gitlab_token(url, token):
    gl = gitlab.Gitlab(url, timeout=10, private_token=token)

    try:
        gl.auth()
    except GitlabAuthenticationError:
        return False

    return True


@click.group(name='login', short_help='authenticate with various services')
@click.pass_obj
def login(ctx):
    pass


@login.command(name='stumpf', help='authenticate with stumpf')
@click.pass_obj
def login_stumpf(ctx):
    click.echo('A browser window has been opened, login and copy the token to login.', err=True)
    webbrowser.open(ctx['stumpf_url'] + '/login?next=/token')

    while True:
        stumpf_token = click.prompt('Token', err=True)
        if _is_valid_stumpf_token(ctx['stumpf_url'], stumpf_token):
            update_config_value(BOILER_CONFIG_FILE, 'stumpf_token', stumpf_token)
            return click.echo(click.style('You are now logged in.', fg='green'), err=True)
        else:
            click.echo(
                click.style(
                    "Your token doesn't appear to be valid, did you copy/paste correctly?",
                    fg='yellow',
                ),
                err=True,
            )


@login.command(name='kwgitlab', help='authenticate with kwgitlab')
@click.pass_obj
def login_kwgitlab(ctx):
    click.echo(
        'A browser window has been opened, login and create a token with the api and \n'
        'read_user scopes and paste it here:',
        err=True,
    )
    webbrowser.open(f"{ctx['gitlab_url']}/profile/personal_access_tokens")

    while True:
        kwgitlab_token = click.prompt('personal access token', err=True)

        if _is_valid_gitlab_token(ctx['gitlab_url'], kwgitlab_token):
            update_config_value(BOILER_CONFIG_FILE, 'kwgitlab_token', kwgitlab_token)
            return click.echo(click.style('You are now authenticated.', fg='green'), err=True)
        else:
            click.echo(
                click.style(
                    "Your token doesn't appear to be valid, did you copy/paste correctly?",
                    fg='yellow',
                ),
                err=True,
            )


cli.add_command(login)



from unittest import mock

import pytest

from boiler import fakes
from boiler.definitions import ActivityType, ActorType
from boiler.models import ActivityList
from boiler.validate import (
    validate_activities,
    validate_activity,
    validate_activity_actor_types,
    validate_actor,
)

negative_tests = [
    ('v', ActivityType.PERSON_TALKS_TO_PERSON),
    ('', ActivityType.PERSON_TALKS_TO_PERSON),
    ('voo', ActivityType.VEHICLE_PICKS_UP_PERSON),
    ('p', ActivityType.PERSON_STEALS_OBJECT),
    ('poo', ActivityType.PERSON_UNLOADS_VEHICLE),
    ('bvb', ActivityType.PERSON_UNLOADS_VEHICLE),
]

positive_tests = [
    ('p', ActivityType.PERSON_TALKS_ON_PHONE),
    ('bopo', ActivityType.PERSON_STEALS_OBJECT),
    ('po', ActivityType.PERSON_STEALS_OBJECT),
    ('p', ActivityType.PERSON_ENTERS_SCENE_THROUGH_STRUCTURE),
    ('vp', ActivityType.PERSON_UNLOADS_VEHICLE),
    ('voooooooopooooooo', ActivityType.PERSON_UNLOADS_VEHICLE),
    ('DEADBEEF', ActivityType.CARRY_IN_HANDS),
]


@pytest.mark.parametrize('input_str,input_type', negative_tests)
def test_validate_activity_actor_string_negative(input_str, input_type):
    warnings, _ = validate_activity_actor_types(input_str, input_type, -1)
    assert len(warnings) == 1


@pytest.mark.parametrize('input_str,input_type', positive_tests)
def test_validate_activity_actor_string_positive(input_str, input_type):
    warnings, errors = validate_activity_actor_types(input_str, input_type, -1)

    assert not warnings
    assert not errors


def test_activity_invalid_combination():
    # incorrect combination of actors
    actor1 = fakes.ActorFactory(begin=0, end=5, actor_type=ActorType.BIKE, detections=[])
    actor2 = fakes.ActorFactory(begin=0, end=10, actor_type=ActorType.PERSON, detections=[])
    activity = fakes.ActivityFactory(
        begin=0,
        end=10,
        activity_type=ActivityType.PERSON_TRANSFERS_OBJECT,
        actors=[actor1, actor2],
    )
    with mock.patch('boiler.validate.validate_actor') as m:
        m.return_value = [], []
        warnings, _ = validate_activity(activity)

    assert 'spec' in str(warnings[0])


def test_validate_actor():
    # valid activity
    d0 = fakes.DetectionFactory(frame=12, keyframe=True)
    d1 = fakes.DetectionFactory(frame=24, keyframe=True)
    actor = fakes.ActorFactory(begin=12, end=24, actor_type=ActorType.VEHICLE, detections=[d0, d1])
    with mock.patch('boiler.validate.validate_detection') as m:
        m.return_value = [], []
        warnings, errors = validate_actor(actor, -1, -1)
        assert not warnings
        assert not errors


def test_valid_actor_detections_outside_activity_range():
    # detection out of range allowed
    d0 = fakes.DetectionFactory(frame=12, keyframe=True)
    d1 = fakes.DetectionFactory(frame=25, keyframe=True)
    actor = fakes.ActorFactory(begin=12, end=24, actor_type=ActorType.VEHICLE, detections=[d0, d1])
    with mock.patch('boiler.validate.validate_detection') as m:
        m.return_value = [], []
        warnings, errors = validate_actor(actor, -1, -1)
        assert not warnings
        assert not errors


def test_valid_actor_detections_outside_activity_range_no_keyframe():
    # detection out of range rejected if not keyframe
    d0 = fakes.DetectionFactory(frame=12, keyframe=True)
    d1 = fakes.DetectionFactory(frame=25, keyframe=False)
    actor = fakes.ActorFactory(begin=12, end=24, actor_type=ActorType.VEHICLE, detections=[d0, d1])
    with mock.patch('boiler.validate.validate_detection') as m:
        m.return_value = [], []
        warnings, errors = validate_actor(actor, -1, -1)
        assert 'last detection must be keyframe' in str(errors[0])


def test_valid_actor_non_monotomic_increasing_detections():
    d0 = fakes.DetectionFactory(frame=12)
    d1 = fakes.DetectionFactory(frame=12)
    actor = fakes.ActorFactory(begin=12, end=12, actor_type=ActorType.VEHICLE, detections=[d0, d1])
    with mock.patch('boiler.validate.validate_detection') as m:
        m.return_value = [], []
        warnings, errors = validate_actor(actor, -1, -1)
        assert 'was not greater than' in str(errors[0])


def test_keyframe_density_too_high():
    detections = [fakes.DetectionFactory(frame=frame, keyframe=True) for frame in range(12, 25)]
    actor = fakes.ActorFactory(
        begin=12, end=24, actor_type=ActorType.VEHICLE, detections=detections
    )
    with mock.patch('boiler.validate.validate_detection') as m:
        m.return_value = [], []
        warnings, errors = validate_actor(actor, -1, -1)
        assert 'keyframe density' in str(warnings[0])


def test_empty_activity_list():
    validate_activities(ActivityList(activity_map={}, actor_map={}))


@pytest.mark.parametrize('begin,end', [(5, 10), (35, 40), (25, 27),])
def test_invalid_actor_frame_range(begin, end):
    actor = fakes.ActorFactory(begin=begin, end=end, actor_type=ActorType.PERSON)
    activity = fakes.ActivityFactory(
        begin=20, end=30, actors=[actor], activity_type=ActivityType.PERSON_TALKS_ON_PHONE
    )
    with mock.patch('boiler.validate.validate_detection') as m:
        with mock.patch('boiler.validate.validate_actor') as m2:
            m.return_value, m2.return_value = ([], []), ([], [])
            warnings, errors = validate_activity(activity)



from functools import wraps
from typing import Callable, Iterable

from boiler.models import ActivityList


SerializeToFile = Callable[[str, Iterable[ActivityList]], None]


class ParseError(Exception):
    pass


def error_handler(message):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ParseError:
                raise
            except Exception as e:
                raise ParseError(message) from e

        return wrapper

    return decorator



import os

from setuptools import find_packages, setup


def prerelease_local_scheme(version):
    """Return local scheme version unless building on master in Gitlab.

    This function returns the local scheme version number
    (e.g. 0.0.0.dev<N>+g<HASH>) unless building on Gitlab for a
    pre-release in which case it ignores the hash and produces a
    PEP440 compliant pre-release version number (e.g. 0.0.0.dev<N>).

    """

    from setuptools_scm.version import get_local_node_and_date

    if 'CI_COMMIT_REF_NAME' in os.environ and os.environ['CI_COMMIT_REF_NAME'] == 'master':
        return ''
    else:
        return get_local_node_and_date(version)


with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='diva-boiler',
    author='Kitware, Inc.',
    author_email='kitware@kitware.com',
    description='a cli for interacting with stumpf server',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(include=['boiler']),
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
        'attrs',
        'boto3',
        'click>=7.0',
        'GitPython',
        'packaging',
        'python-gitlab',
        'pyxdg',
        'pyyaml',
        'requests',
        'requests-toolbelt',
        'sentry-sdk',
        'tabulate',
    ],
    entry_points={'console_scripts': ['boiler=boiler:main']},
    license='Apache Software License 2.0',
    use_scm_version={
        'local_scheme': prerelease_local_scheme,
        'root': '..',
        'relative_to': __file__,
    },
    setup_requires=['setuptools_scm'],
)
